@@ -1,5 +1,4 @@
 import json
-import logging
 import os
 import time
 from enum import Flag
@@ -8,19 +7,20 @@ from threading import Event
 from pkg_resources import get_distribution
 
 import requests
-from attrdict import AttrDict
 
-from .exceptions import HttpError, NetworkError
+from .exceptions import HttpError, NetworkError, SmallDError
 from .gateway import Gateway
+from .json_elements import JsonObject
+from .logger import logger
 from .ratelimit import RateLimiter
 from .standard_listeners import add_standard_listeners
 
-logger = logging.getLogger("smalld")
-
-
 __version__ = get_distribution("smalld").version
 
 
+MIN_SECONDS_BETWEEN_CONNECTIONS = 120
+
+
 class Intent(Flag):
     GUILDS = 1 << 0
     GUILD_MEMBERS = 1 << 1
@@ -58,24 +58,33 @@ recoverable_error_codes = {
 }
 
 
+def is_recoverable_error(reason):
+    if not reason or not reason.code:
+        return True
+
+    return reason.code in recoverable_error_codes
+
+
 class SmallD:
     def __init__(
         self,
         token=os.environ.get("SMALLD_TOKEN"),
         base_url="https://discord.com/api/v6",
         intents=Intent.unprivileged(),
+        shard=(0, 1),
     ):
         if not token:
-            raise ValueError("No bot token provided")
+            raise SmallDError("No bot token provided")
 
         self.token = token
         self.base_url = base_url
         self.intents = intents
+        self.shard = shard
+
         self.listeners = []
         self.closed_event = Event()
 
         self.http = HttpClient(token, base_url)
-
         self.get = self.http.get
         self.post = self.http.post
         self.put = self.http.put
@@ -86,20 +95,21 @@ class SmallD:
 
     def __getattr__(self, name):
         if name.startswith("on_"):
-            return lambda: self.on_dispatch(t=name.strip("on_").upper())
+            return lambda func=None: self.on_dispatch(func, t=name.strip("on_").upper())
 
         super().__getattr__(name)
 
-    def on_dispatch(self, t):
+    def on_dispatch(self, func=None, *, t=None):
         def decorator(f):
-            self.on_gateway_payload(op=0, t=t)(lambda payload: f(payload.d))
+            self.on_gateway_payload(lambda payload: f(payload.d), op=0, t=t)
+            return f
 
-        return decorator
+        return decorator if func is None else decorator(func)
 
-    def on_gateway_payload(self, op=None, t=None):
+    def on_gateway_payload(self, func=None, *, op=None, t=None):
         def decorator(f):
             def filtered_payload_listener(data):
-                if op and data.op != op:
+                if op is not None and data.op != op:
                     return
 
                 if t and data.t != t:
@@ -108,13 +118,12 @@ class SmallD:
                 f(data)
 
             self.listeners.append(filtered_payload_listener)
+            return f
 
-        return decorator
+        return decorator if func is None else decorator(func)
 
     def send_gateway_payload(self, data):
-        payload = json.dumps(data)
-        logger.debug("gateway payload sent: %s", payload)
-        self.gateway.send(payload)
+        self.gateway.send(data)
 
     @property
     def closed(self):
@@ -142,25 +151,36 @@ class SmallD:
 
         while not self.closed:
             logger.info("Gateway connecting...")
-            gateway_url = self.get("/gateway/bot").url
+            connection_time = int(time.monotonic())
 
-            self.gateway = Gateway(gateway_url)
+            try:
+                gateway_url = self.get("/gateway/bot").url
+            except (HttpError, NetworkError) as e:
+                logger.info(f"Could not fetch gateway url. ({type(e).__name__}) {e}")
+            else:
+                self.gateway = Gateway(gateway_url)
 
-            for data in self.gateway:
-                logger.debug("gateway payload received: %s", data)
-                for listener in self.listeners:
-                    listener(data)
+                for data in self.gateway:
+                    self.notify_listeners(data)
 
-            reason = self.gateway.close_reason
-            if reason and not reason.code in recoverable_error_codes:
-                logger.fatal(
-                    "Unrecoverable gateway closure: %s", self.gateway.close_reason
-                )
-                self.close()
+                if not is_recoverable_error(self.gateway.close_reason):
+                    logger.fatal(
+                        "Unrecoverable gateway closure: %s", self.gateway.close_reason
+                    )
+                    self.close()
 
             if not self.closed:
-                logger.info("Pausing before reconnect...")
-                time.sleep(5)
+                since_last_connection = int(time.monotonic()) - connection_time
+                time.sleep(
+                    max(5, MIN_SECONDS_BETWEEN_CONNECTIONS - since_last_connection)
+                )
+
+    def notify_listeners(self, data):
+        try:
+            for listener in self.listeners:
+                listener(data)
+        except:
+            logger.warn("Exception in listener", exc_info=True)
 
 
 class HttpClient:
@@ -223,7 +243,7 @@ class HttpClient:
         except json.JSONDecodeError:
             raise HttpError(response=res)
 
-        return AttrDict(content)
+        return JsonObject(content)
 
     def close(self):
         self.session.close()
